// Glove controller v3
//
// Changelog:
//  2015-05-19:
//    v3 - slower comms blinking (to be noticeable)
//    v3 - compiled with a new TaskScheduler version 1.4 (sleep IDLE instead of delay)
//  2015-05-20:
//    v3a - change base rotation control to be accel driven as well
//    v3a - complied against new TaskScheduler with IDLE sleep 


// INCLUDES

#include <DirectIO.h>

#include <I2Cdev.h>
#include <Wire.h>
#include <MPU6050.h>
#include <DigitalFilter.h>

#include <SPI.h>
#include "nRF24L01.h"
#include "RF24.h"

#include <TaskScheduler.h>


// TEST AND DEBUG DEFINES

//#define _DEBUG_
//#define _TEST_


// DEFINES

#define  CLOSE_CLAW_BUTTON  4
#define  OPEN_CLAW_BUTTON  3
#define  FLEX_SENSOR  A0
#define  LED  A1
#define  CE_PIN   9
#define  CSN_PIN  10

Input<CLOSE_CLAW_BUTTON> pCloseClaw;
Input<OPEN_CLAW_BUTTON>  pOpenClaw;
Output<LED>              pLed;
AnalogInput<FLEX_SENSOR> pFlex;

#define M_OFF      0b00
#define M_ON        0b11
#define M_RIGHT   0b01
#define M_LEFT     0b10

#define ACCON_TH  9000
#define ACCOFF_TH 8000

#define FLIP_TH  (-3000)
#define GYRO_TH   4000

#define  RADIO_DEFAULT_CHANNEL  56

// STRUCTURES

struct packed_bits {
  byte m1  : 2; // claw
  byte m2  : 2; // wrist
  byte m3  : 2; // elbow
  byte m4  : 2; // arm
  byte m5  : 2; // base rotate
  byte led : 2; // claw LED
} ctrlWord, cwTx, cwPrev;


// TASK DEFINES

#define MEASURE_PERIOD  50   // 20 times/sec
#define TRANSMIT_PERIOD 100   // 2 times/sec or on demand
#define TRAN_LED_PERIOD 200  // 5 times/sec
#define REACTION_DELAY  100   //

#define FLEX_OPN_ON   450
#define FLEX_OPN_OFF  440
#define FLEX_CLS_ON   390
#define FLEX_CLS_OFF  400


// TASKS

Scheduler runner;

Task tMeasure(MEASURE_PERIOD, -1, &measureCallback);
Task tLedDelay(REACTION_DELAY, 1, &ledDelayCallback);
Task tB1Delay(REACTION_DELAY, 1, &b1DelayCallback);
Task tB2Delay(REACTION_DELAY, 1, &b2DelayCallback);
Task tFlexDelay(REACTION_DELAY, -1, &flexDelayCallback);
Task tTransmit(TRANSMIT_PERIOD, -1, &transmitCallback);

#ifdef _DEBUG_
Task tDisplay(1000, -1, &displayCallback);
#endif


// VARIABLES

const uint64_t pipe = 0xE8E8F0F0E1LL; // Define the transmit pipe
volatile bool rts;
volatile bool radio_status;

// Glove controls/sensors


// DRIVERS

MPU6050 ag;
RF24 radio(CE_PIN, CSN_PIN);


// CODE: CALLBACKS

int b1, b2, flex, ax, ay, az, gx, gy, gz;
avgFilter axF(5), ayF(5), azF(5);
dhpFilter gxF, gyF, gzF;

void measureCallback() {

  bool changed = false;

  cwPrev = ctrlWord;
  
  ag.getMotion6(&ax, &ay, &az, &gx, &gy, &gz);
  ax = axF.value(ax);
  ay = ayF.value(ay);
  az = azF.value(az);
  gx = gxF.value(gx);
  gy = gyF.value(gy);
//  gz = gzF.value(gz);
  
  b1 = pCloseClaw;
  b2 = pOpenClaw;

  // Check if we need to flip the led (both buttons pressed)
  if (b1 && b2) {
    if ( !tLedDelay.isEnabled() ) {
      tLedDelay.enableDelayed();
      tLedDelay.setIterations(1);
    }
    tB1Delay.disable();
    tB2Delay.disable();
    ctrlWord.m1 = M_OFF;
  }
  else {
    tLedDelay.disable();

    if (b1) {
      if ( !tB1Delay.isEnabled() ) {
        tB1Delay.enableDelayed();
        tB1Delay.setIterations(1);
      }
    }

    if (b2) {
      if ( !tB2Delay.isEnabled() ) {
        tB2Delay.enableDelayed();
        tB2Delay.setIterations(1);
      }
    }

    if ( !(b1 || b2) ) {
      ctrlWord.m1 = M_OFF;
      tB1Delay.disable();
      tB2Delay.disable();
    }
  }

  // Flex
  flex = pFlex;
  if (flex > FLEX_OPN_OFF) {
    ctrlWord.m2 = M_OFF;
    tFlexDelay.disable();
  }
  else {
    if ( !tFlexDelay.isEnabled() ) tFlexDelay.enableDelayed();
  }


  // Elbow
  if (ax > ACCON_TH && abs(gx) < GYRO_TH) {
    ctrlWord.m3 = M_RIGHT;
  }
  else if (ax < -ACCON_TH && abs(gx) < GYRO_TH) {
    ctrlWord.m3 = M_LEFT;
  }
  else if (abs(ax) < ACCOFF_TH) {
    ctrlWord.m3 = M_OFF;
  }


  // Arm or Base
  if (ay > ACCON_TH && abs(gy) < GYRO_TH) {
    if (az > 0) {
      ctrlWord.m4 = M_RIGHT;
      ctrlWord.m5 = M_OFF;
    }
    if (az < FLIP_TH) {
      ctrlWord.m5 = M_LEFT;
      ctrlWord.m4 = M_OFF;
    }
  }
  else if (ay < -ACCON_TH && abs(gy) < GYRO_TH) {
    if (az > 0) {
      ctrlWord.m4 = M_LEFT;
      ctrlWord.m5 = M_OFF;
    }
    if (az < FLIP_TH) {
      ctrlWord.m5 = M_RIGHT;
      ctrlWord.m4 = M_OFF;
    }
  }
  else if (abs(ay) < ACCOFF_TH) {
    ctrlWord.m4 = M_OFF;
    ctrlWord.m5 = M_OFF;
  }

  if ( !cwEqual(cwPrev, ctrlWord) ) tTransmit.enable();
}


void flexDelayCallback() {
  if (flex < FLEX_OPN_ON && flex >= FLEX_CLS_OFF) {
    ctrlWord.m2 = M_LEFT;
    tTransmit.enable();
  }

  if (flex < FLEX_CLS_ON) {
    ctrlWord.m2 = M_RIGHT;
    tTransmit.enable();
  }
}


void ledDelayCallback() {
  ctrlWord.led = (ctrlWord.led == M_OFF) ? M_ON : M_OFF;
  tTransmit.enable();
}


void b1DelayCallback() {
  ctrlWord.m1 = M_LEFT;
  tTransmit.enable();
}


void b2DelayCallback() {
  ctrlWord.m1 = M_RIGHT;
  tTransmit.enable();
}



void transmitCallback() {

  if (!rts) return;

  noInterrupts();
  cwTx = ctrlWord;
  rts = false;
  interrupts();

  if (radio_status) pLed = HIGH;
  radio.startWrite( &cwTx, sizeof(cwTx) );
}


#ifdef _DEBUG_
void displayCallback() {
  char  ln[256];
  byte *p;

  p = (byte *) &ctrlWord;

  sprintf(ln, "led=%02d\tm1=%02d\tm2=%02d\tm3=%02d\tm4=%02d\tm5=%02d\tgDir=%08d\tflex=%04d", ctrlWord.led, ctrlWord.m1, ctrlWord.m2, ctrlWord.m3, ctrlWord.m4, ctrlWord.m5, gDir, flex);
  Serial.println(ln);
  Serial.print(*p, HEX); p++; Serial.print(":"); Serial.println(*p, HEX);
}
#endif



// CODE: MAIN
void initCtrlWord() {
  ctrlWord.led = M_OFF;
  ctrlWord.m1 = M_OFF;
  ctrlWord.m2 = M_OFF;
  ctrlWord.m3 = M_OFF;
  ctrlWord.m4 = M_OFF;
  ctrlWord.m5 = M_OFF;
}


void initPins() {
  //  pinMode(CLOSE_CLAW_BUTTON, INPUT);
  //  pinMode(OPEN_CLAW_BUTTON, INPUT);
  //  pinMode(FLEX_SENSOR, INPUT);
  //  pinMode(LED, OUTPUT);
  pinMode(10, OUTPUT); // SPI needs it?
}


void initRadio() {

  //  radio.setChannel(RADIO_DEFAULT_CHANNEL);  // 0-127
  radio.setDataRate(RF24_250KBPS);  // RF24_1MBPS = 0, RF24_2MBPS, RF24_250KBPS
  //  radio.setRetries(5,5);
  radio.setPayloadSize( sizeof(ctrlWord) );
  //  radio.setAutoAck(true);
  //  radio.enableAckPayload();
  radio.setCRCLength(RF24_CRC_8);  // RF24_CRC_DISABLED = 0, RF24_CRC_8, RF24_CRC_16
  //  radio.setPALevel(RF24_PA_HIGH); // RF24_PA_MIN=-18dBm, RF24_PA_LOW=-12dBm, RF24_PA_MED=-6dBM, and RF24_PA_HIGH=0dBm. *  RF24_PA_MIN = 0,RF24_PA_LOW, RF24_PA_HIGH, RF24_PA_MAX, RF24_PA_ERROR
  delay(10);
  //  radio.testCarrier();
}



void setup() {
  // put your setup code here, to run once:
  initPins();
  initCtrlWord();

  Wire.begin();       //Initiate the Wire library and join the I2C bus as a master

#ifdef _DEBUG_
  Serial.begin(115200);
  Serial.println("Robotic Arm Glove Controller Circuit test");
#endif

#ifndef _DEBUG_
  power_usart0_disable();
#endif

  // Configure radio
  radio.begin();
  initRadio();

  ag.initialize();
  if (!ag.testConnection()) {
#ifdef _DEBUG_
    Serial.println("Accelerometer is not responding");
#endif
    while (1) {};
  }

  attachInterrupt(0, check_radio, FALLING);
  rts = true;

  runner.init();

  runner.addTask(tMeasure);
  runner.addTask(tLedDelay);
  runner.addTask(tB1Delay);
  runner.addTask(tB2Delay);
  runner.addTask(tFlexDelay);
  runner.addTask(tTransmit);
  
#ifdef _DEBUG_
  runner.addTask(tDisplay);
  tDisplay.enable();
#endif

  tMeasure.enable();
  tTransmit.enable();

  for (int i = 1; i < 6; i++) {
    pLed = i & 1 ? HIGH : LOW;
    delay(400);
  }



  // Start transmitting
  radio.powerUp();

  for (int i = 0; i < 12; i++) {
    pLed = i & 1 ? HIGH : LOW;
    delay(200);
  }

  radio.openWritingPipe(pipe);
}


void loop() {
  runner.execute();
}


void check_radio(void)
{
  // What happened?
  bool tx, fail, rx;
  radio.whatHappened(tx, fail, rx);
  
  pLed = LOW;
  radio_status = false;
  // Have we successfully transmitted?
  if ( tx )
  {
    radio_status = true;
  }

  rts = true;
}

bool cwEqual(struct packed_bits& a, struct packed_bits& b) {
  return ( (a.led == b.led) && (a.m1 == b.m1) && (a.m2 == b.m2) && (a.m3 == b.m3) && (a.m4 == b.m4) && (a.m5 == b.m5) );
}

